/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Automatically takes a snapshot of the state of an IAutoRecoverable
 * 			object before it is modified
 * 			
 */
package ca.mcgill.acid.aspects;
 
import ca.mcgill.acid.interfaces.IAutoRecoverable;
import ca.mcgill.acid.interfaces.IRecoverable;
import ca.mcgill.acid.interfaces.Constants;
import ca.mcgill.acid.interfaces.IAccessClassified;
import ca.mcgill.acid.utility.RecoveryZone;
import java.util.Vector;
import ca.mcgill.acid.utility.View;
import ca.mcgill.acid.utility.NoZoneFoundForThisThreadException;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import java.util.StringTokenizer;
import ca.mcgill.acid.utility.Version;
 
public aspect AutoRecoverable {
   
	declare parents: IAutoRecoverable implements IRecoverable,IAccessClassified; 
	 
	private static InheritableThreadLocal myView = new InheritableThreadLocal();
	
	/**
     * Method execution on the target object IAutoRecoverable
     */
	pointcut methodExecution( IAutoRecoverable autoRecover ): 
		target(autoRecover) && call(public * IAutoRecoverable+.*(..));
	  
	/**
	 * Take a snapshot of the target object
	 */
	before(IAutoRecoverable autoRecover):  methodExecution(autoRecover) && if(myView.get()!= null)
	{
	  try{
			if(!Constants.READ.equalsIgnoreCase(autoRecover.getAccessType(getMethodName(thisJoinPoint.toShortString()))))
			{
				if(((RecoveryZone)myView.get()).getBeginZone()
				&& !objectInMyZone(((RecoveryZone)myView.get()).getZone(), autoRecover.getMyName()))
				{
					int id = autoRecover.establishCheckpoint(((RecoveryZone)myView.get()).getThreadView());
					Version v = new Version(id,autoRecover.getMyName());
					((RecoveryZone)myView.get()).getZone().add(v);
				}
			}
		}
		catch (MethodNotAnnotatedException e)
		{e.printStackTrace();}
		catch (MethodNotFoundException e)
		{e.printStackTrace();}
		
	}
	
	/**
	 * Undo all the changes made by the invoking thread
	 */
	public static synchronized void undo()
	{
		if(myView.get()!= null)
		{
			View view = ((RecoveryZone)myView.get()).getThreadView();
			Vector zone = ((RecoveryZone)myView.get()).getZone();
			
			for(int i = 0; i < view.getObjectsInView().size(); i++ )
			{
				for(int j = 0; j < zone.size(); j++)
				{
					if(((Version)view.getObjectsInView().elementAt(i)).getObjectName().equals(
					   ((Version)zone.elementAt(i)).getObjectName()))
					  {
						  ((Version)view.getObjectsInView().elementAt(i)).setVersionID(
								  ((Version)zone.elementAt(i)).getVersionID());
					  }
				}
			  
			}
		}
	}
	
	/**
	 * Begin a new recoverable zone for the invoking thread
	 */
	public static synchronized void beginRecoverableZone(View view)
	{
		if(myView.get()== null)
		{
			Vector zone = new Vector();
			RecoveryZone recoveryZone = new RecoveryZone(view,zone);
			myView.set(recoveryZone);
		}
	}

	/**
	 * End the recoverable zone of the invoking thread
	 */
	public static synchronized void endRecoverableZone()
		throws NoZoneFoundForThisThreadException
	{
		if(myView.get()!= null)
			((RecoveryZone)myView.get()).setBeginZone(false);
		else
			throw new NoZoneFoundForThisThreadException(
					"\n	No Zone found for thread <" +Thread.currentThread().getName()+ "> \n");
	}
	
	/**
	 * Join an existing recoverable zone
	 */
	public static synchronized void joinRecoverableZone(Vector zone)
		throws NoZoneFoundForThisThreadException
	{
		if(myView.get()!= null)
			((RecoveryZone)myView.get()).setZone(zone);
		else
			throw new NoZoneFoundForThisThreadException(
					"\n	No Zone found for thread <" +Thread.currentThread().getName()+ "> \n");
	}
	
	/**
	 * Leave your current recoverable zone
	 */
	public static synchronized void leaveRecoverableZone()
		throws NoZoneFoundForThisThreadException
	{
		if(myView.get()!= null)
			((RecoveryZone)myView.get()).setZone(null);
		else
			throw new NoZoneFoundForThisThreadException(
				"\n	No Zone found for thread <" +Thread.currentThread().getName()+ "> \n");
	}

	
		 
	/**
	 * Returns all the objects automatically created for the invoking thread
	 */
	public static synchronized Vector getObjects()
		throws NoZoneFoundForThisThreadException
	{
		Vector zone = null;
		
		if(myView.get()!= null)
			zone = ((RecoveryZone)myView.get()).getZone();
		else
			throw new NoZoneFoundForThisThreadException(
				"\n	No Zone found for thread <" +Thread.currentThread().getName()+ "> \n");
	
		return zone;
	}

	
	
	/**
	 * Helper method
	 */
	private synchronized boolean objectInMyZone(Vector zone, String objName)
	{
		boolean found = false;
		
		if(zone == null)
			return found;
		
		for(int i = 0; i < zone.size(); i++ )
		{
		  if(((Version)zone.elementAt(i)).getObjectName().equals(objName))
		  {
			  found = true;
			  break;
		  }
		}
		
		return found;
	}
	
	/**
	 * Helper method
	 */
	private synchronized String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
} 