/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Provides Lock-Based functionality to all objects of classes that
 * 			implements the ILockBased interface.
 */

package ca.mcgill.acid.aspects;

import ca.mcgill.acid.utility.Lock;
import ca.mcgill.acid.interfaces.ILockBased;
import ca.mcgill.acid.interfaces.IAccessClassified;
import ca.mcgill.acid.interfaces.IAutoRecoverable;
import ca.mcgill.acid.interfaces.ITracked;
import ca.mcgill.acid.interfaces.IShared;
import ca.mcgill.acid.utility.TransactionLocks;
import java.util.StringTokenizer;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import ca.mcgill.acid.utility.NoZoneFoundForThisThreadException;
import ca.mcgill.acid.interfaces.Constants;

public aspect LockBased {
	
	declare parents: 
		ILockBased implements IAccessClassified,IShared,IAutoRecoverable,ITracked;

	// Keeps track of the types of locks a transaction acquire for each
	// LockBased object it accesses. it will facilitate the release of
	// the appropriate locks at commit/abort time.
	private static InheritableThreadLocal thisLocks = new InheritableThreadLocal();

	private Lock ILockBased.transactionLock = new Lock();
	private Lock ILockBased.getLock()  { return transactionLock; }
	   
	/**
	 * Method call on the target object ILockBased
	 */
	pointcut methodCall( ILockBased lockBased ): 
		target(lockBased) && call(public * ILockBased+.*(..));
	
	/**
	 * Joinpoints that signals transaction termination and a request to release
	 * transactional locks 
	 */
	pointcut signalTransactionTermination(): 
		call(public static void commitLockBased()) || 
		call(public static void abortLockBased());
	
	/**
	 * Synchronizes access to the data members of ILockBased objects
	 * by ensuring that conflicting operations are not executed
	 * concurrently by Transactions
	 */
	Object around(ILockBased lockBased) :  methodCall(lockBased)
	{
		// set ThreadLocal once per transaction
		if(thisLocks.get()== null)
		{
			TransactionLocks myLocks = new TransactionLocks();
			thisLocks.set(myLocks);
		}
		
		Object obj;
		String accessType = null;
		
		//get access type
		try{
			accessType = lockBased.getAccessType(getMethodName(thisJoinPoint.toShortString()));
			
		}
		catch (MethodNotAnnotatedException e)
		{e.printStackTrace();}
		catch (MethodNotFoundException e)
		{e.printStackTrace();}
		
		
		//get appropriate transactional lock
		if(Constants.READ.equalsIgnoreCase(accessType)){
			if(!Thread.currentThread().getName().equals(lockBased.getLock().getWriteLockHolder())
			&& !Thread.currentThread().getName().equals(lockBased.getLock().getUpdateLockHolder())
			&& !Thread.currentThread().getName().equals(lockBased.getLock().getReadLockHolder()))
				{
				 lockBased.getLock().getReadLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
				 ((TransactionLocks)thisLocks.get()).addReadLock(lockBased.getMyName());
				 System.out.println("[Get_Transaction_Read_Lock @ <"+ lockBased.getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
				}
			
		}

		if(Constants.WRITE.equalsIgnoreCase(accessType)){
			if(!Thread.currentThread().getName().equals(lockBased.getLock().getWriteLockHolder())
			&& !Thread.currentThread().getName().equals(lockBased.getLock().getUpdateLockHolder()))	
			{
				if(Thread.currentThread().getName().equals(lockBased.getLock().getReadLockHolder()))
				{
					  lockBased.getLock().releaseReadLock();
					  ((TransactionLocks)thisLocks.get()).getReadLocks().remove(lockBased.getMyName());
					  System.out.println("[Release_Transaction_Read_Lock @ <"+ lockBased.getMyName() + "> For_Thread <" +Thread.currentThread().getName() +">");
				}
				lockBased.getLock().getWriteLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
				 ((TransactionLocks)thisLocks.get()).addWriteLock(lockBased.getMyName());
				 System.out.println("[Get_Transaction_Write_Lock @ <"+ lockBased.getMyName() + "> For_Thread <" +Thread.currentThread().getName() +">");
			}
		}
		 
		if(Constants.UPDATE.equalsIgnoreCase(accessType)) {
			
			if(!Thread.currentThread().getName().equals(lockBased.getLock().getWriteLockHolder())
			&& !Thread.currentThread().getName().equals(lockBased.getLock().getUpdateLockHolder()))					
			{	
				if(Thread.currentThread().getName().equals(lockBased.getLock().getReadLockHolder()))
					{
					  lockBased.getLock().releaseReadLock();
					  ((TransactionLocks)thisLocks.get()).getReadLocks().remove(lockBased.getMyName());
					  System.out.println("[Release_Transaction_Read_Lock @ <"+ lockBased.getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
					}
				
				lockBased.getLock().getUpdateLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
				 ((TransactionLocks)thisLocks.get()).addUpdateLock(lockBased.getMyName());
				 System.out.println("[[ Get_Transaction_Update_Lock @ <"+ lockBased.getMyName() + "> For_Thread <" +Thread.currentThread().getName() + ">");
			}  
			
		} 
		
		
		// -> select in-place update <-
		 lockBased.setDeferred(false);
		
		obj = proceed(lockBased);
		
		return obj;
	}
	
	/**
	 *  Release transactional locks
	 */
	after() : signalTransactionTermination()
	{
		try{
			for(int j = 0; j <  ((TransactionLocks)thisLocks.get()).getReadLocks().size(); j++)
			{ 
				for(int i = 0; i < Tracked.getReadObjects().size(); i++)
				{	
					if(Tracked.getReadObjects().elementAt(i) instanceof ILockBased )
					{
						if( ((String)((TransactionLocks)thisLocks.get()).getReadLocks().elementAt(j)).equals(
								((ILockBased)Tracked.getReadObjects().elementAt(i)).getMyName()))
						{
							System.out.println("[Release_Transaction_Read_Lock @ <"+ 
									((ILockBased)Tracked.getReadObjects().elementAt(i)).getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
							((ILockBased)Tracked.getReadObjects().elementAt(i)).getLock().releaseReadLock();
						
						}
					} 
						
				} 
			}
			
			for(int j = 0; j <  ((TransactionLocks)thisLocks.get()).getWriteLocks().size(); j++)
			{
				for(int i = 0; i < Tracked.getWriteObjects().size(); i++)
				{
					if(Tracked.getWriteObjects().elementAt(i) instanceof ILockBased )
					{
						if( ((String)((TransactionLocks)thisLocks.get()).getWriteLocks().elementAt(j)).equals(
							((ILockBased)Tracked.getWriteObjects().elementAt(i)).getMyName()))
						{
							System.out.println("[Release_Transaction_Write_Lock @ <"+ 
									((ILockBased)Tracked.getWriteObjects().elementAt(i)).getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
							((ILockBased)Tracked.getWriteObjects().elementAt(i)).getLock().releaseWriteLock();
						}
					}
						
				}
			}
			
			for(int j = 0; j <  ((TransactionLocks)thisLocks.get()).getUpdateLocks().size(); j++)
			{
				for(int i = 0; i < Tracked.getUpdatedObjects().size(); i++)
				{
					if(Tracked.getUpdatedObjects().elementAt(i) instanceof ILockBased )
					{
						if( ((String)((TransactionLocks)thisLocks.get()).getUpdateLocks().elementAt(j)).equals(
							((ILockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName()))
						{
							System.out.println("[Release_Transaction_Update_Lock @ <"+ 
									((ILockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName() + "> For_Thread <" +Thread.currentThread().getName() +">");
							((ILockBased)Tracked.getUpdatedObjects().elementAt(i)).getLock().releaseUpdateLock();
						}
					}
						
				}
			}
			
			
		}catch(NoZoneFoundForThisThreadException e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Helper method to around advice
	 */
	private synchronized String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
	  

}
