
/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Provides Multi-Version Lock-Based functionality to all objects of classes that
 * 			implements the ILockBased interface.
 */

package ca.mcgill.acid.aspects;

import ca.mcgill.acid.utility.Lock;
import ca.mcgill.acid.utility.TransactionLocks;
import ca.mcgill.acid.utility.Version;
import ca.mcgill.acid.interfaces.IMultiVersionLockBased;
import ca.mcgill.acid.interfaces.IAccessClassified;
import ca.mcgill.acid.interfaces.IAutoRecoverable;
import ca.mcgill.acid.interfaces.ITracked;
import ca.mcgill.acid.interfaces.IShared;
import ca.mcgill.acid.utility.NoCreationTimeStampException;
import ca.mcgill.acid.utility.NoViewFoundException;
import java.util.StringTokenizer;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import ca.mcgill.acid.utility.NoZoneFoundForThisThreadException;
import ca.mcgill.acid.interfaces.Constants;
import ca.mcgill.acid.utility.GlobalCounter;
import java.util.Vector;

public aspect MultiVersionLockBased {
	
	declare parents: 
		IMultiVersionLockBased implements IAccessClassified,IShared,IAutoRecoverable,ITracked;
	
	//holds the creation timestamp for read-only transactions
	private static InheritableThreadLocal creationTime = new InheritableThreadLocal();

	// Keeps track of the types of locks a transaction acquire for each
	// LockBased object it accesses. it will facilitate the release of
	// the appropriate locks at commit/abort time.
	private static InheritableThreadLocal locks = new InheritableThreadLocal();

	private Lock IMultiVersionLockBased.transactionLock = new Lock();
	private Lock IMultiVersionLockBased.getLock()  { return transactionLock; }
	private Vector historyOfCommitedStates = new Vector();
	private static GlobalCounter counter = new GlobalCounter();
	
	/**
     * Constructor execution on the target object IMultiVersionLockBased
     */
    pointcut constructorExecution(IMultiVersionLockBased multiVersioned):
    	target(multiVersioned) && execution(public IMultiVersionLockBased+.new(..));
	
	
	/**
	 * Method Execution on the target object IMultiVersionLockBased
	 */
	pointcut methodExecution( IMultiVersionLockBased multiVersion ): 
		target(multiVersion) && call(public * IMultiVersionLockBased+.*(..));
	
	/**
	 * Joinpoints that signals transaction commit
	 */
	pointcut signalMultiVersionTransactionCommit(): 
		call(public static void commitMultiVersionLockBased());
	
	
	/**
	 * Joinpoints that signals transaction abort and a request to release
	 * transactional locks 
	 */
	pointcut signalMultiVersionTransactionAbort(): 
		call(public static void abortMultiVersionLockBased());
	
	
	 after(IMultiVersionLockBased multiVersion) returning:
		 constructorExecution(multiVersion)
	    { 
		 Version v = multiVersion.newVersion();
		 v.setCommitTime(counter.getNext());
		 historyOfCommitedStates.add(v);	
	    }
	
	
	/**
	 * Synchronizes access to the data members of ILockBased objects
	 * by ensuring that conflicting operations are not executed
	 * concurrently by Transactions
	 */
	Object around(IMultiVersionLockBased multiVersion) 
	:  methodExecution(multiVersion)
	{
		
		Object obj = null;
		String accessType = null;
		
	   try{
			
			// set ThreadLocal once per transaction
			if(locks.get()== null)
			{
				TransactionLocks myLocks = new TransactionLocks();
				locks.set(myLocks);
			}
			
			//get access type
			accessType = multiVersion.getAccessType(getMethodName(thisJoinPoint.toShortString()));
			
			// Handle update transactions
			if(!Constants.READ.equalsIgnoreCase(accessType)){
				
				//get appropriate transactional lock
				if(Constants.WRITE.equalsIgnoreCase(accessType)){
					if(!Thread.currentThread().getName().equals(multiVersion.getLock().getWriteLockHolder())
					&& !Thread.currentThread().getName().equals(multiVersion.getLock().getUpdateLockHolder()))	
					{
						if(Thread.currentThread().getName().equals(multiVersion.getLock().getReadLockHolder()))
						{
							multiVersion.getLock().releaseReadLock();
							((TransactionLocks)locks.get()).getReadLocks().remove(multiVersion.getMyName());
							System.out.println("[Release_Transaction_Read_Lock @ <"+ multiVersion.getMyName() + "> For_Thread <" +Thread.currentThread().getName() +">");
						}
						multiVersion.getLock().getWriteLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
						 ((TransactionLocks)locks.get()).addWriteLock(multiVersion.getMyName());
						 System.out.println("[Get_Transaction_Write_Lock @ <"+ multiVersion.getMyName() + "> For_Thread <" +Thread.currentThread().getName() +">" );
					}
				}
				 
				if(Constants.UPDATE.equalsIgnoreCase(accessType)) {
					
					if(!Thread.currentThread().getName().equals(multiVersion.getLock().getWriteLockHolder())
					&& !Thread.currentThread().getName().equals(multiVersion.getLock().getUpdateLockHolder()))					
					{	
						if(Thread.currentThread().getName().equals(multiVersion.getLock().getReadLockHolder()))
						{
						  multiVersion.getLock().releaseReadLock();
						  ((TransactionLocks)locks.get()).getReadLocks().remove(multiVersion.getMyName());
						  System.out.println("[Release_Transaction_Read_Lock @ <"+ multiVersion.getMyName() + "> For_Thread <" +Thread.currentThread().getName() +">");
						}
					
						 multiVersion.getLock().getUpdateLock(Thread.currentThread().getName(),getMethodName(thisJoinPoint.toShortString()));
						 ((TransactionLocks)locks.get()).addUpdateLock(multiVersion.getMyName());
						 System.out.println("[Get_Transaction_Update_Lock @ <"+ multiVersion.getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
					
					}  
				} 
				
				// -> select in-place update <-
				multiVersion.setDeferred(false);
			}
			//Handle Read-Only transactions
			else{
					
				//look through history of commited states to find the version with the highest
				//time-stamp that is lower than the Read-Only transaction creation time-stamp
				 if(creationTime.get()!= null)
				 {
					 int transactionCreationTime = ((Integer) (creationTime.get())).intValue();
					 int versionID = -1;
					 int max = -1;
					 
					 for(int i = 0; i < historyOfCommitedStates.size(); i++ )
						{
							if(((Version)historyOfCommitedStates.elementAt(i)).getObjectName().equals(multiVersion.getMyName()))
							{
								if((transactionCreationTime > ((Version)historyOfCommitedStates.elementAt(i)).getCommitTime())
								&& ((Version)historyOfCommitedStates.elementAt(i)).getCommitTime() > max)
								{
									max = ((Version)historyOfCommitedStates.elementAt(i)).getCommitTime();
									versionID = ((Version)historyOfCommitedStates.elementAt(i)).getVersionID();
								}
							}
						}
					 
					 //Assign the version found in the history of commited states to the view of
					 //this Read-only transaction 
					 if(versionID != -1)
					 {
						 try{
						  Versioned.setCurrentVersion(multiVersion.getMyName(), versionID);
					    }catch(NoViewFoundException e)
						{
							e.printStackTrace();
						}
					 }
					 
					 //System.out.println("	View @"+Thread.currentThread().getName()+" : " + multiVersion.getMyName() +" Version: "+ versionID);
				 }
				 else
					 throw new NoCreationTimeStampException("\n The Read-Only transaction <"+
							 Thread.currentThread().getName()+" must request a Creation time by invoking MultiVersionLockBased.startReadOnlyTransaction()");
			}
			
			obj = proceed(multiVersion);
			
		}		
		catch (MethodNotAnnotatedException e)
		{e.printStackTrace();}
		catch (MethodNotFoundException e)
		{e.printStackTrace();}
		catch (NoCreationTimeStampException e)
		{e.printStackTrace();}
		

		return obj;
	}
	
	
	/**
	 *  Assigns timestamps to updated transaction and save their states before Releasing transactional locks
	 */
	after() : signalMultiVersionTransactionCommit()
	{
		try{
			int commitTime = counter.getNext();
			
			for(int j = 0; j <  ((TransactionLocks)locks.get()).getWriteLocks().size(); j++)
			{
				for(int i = 0; i < Tracked.getWriteObjects().size(); i++)
				{
					if(Tracked.getWriteObjects().elementAt(i) instanceof IMultiVersionLockBased )
					{
						if( ((String)((TransactionLocks)locks.get()).getWriteLocks().elementAt(j)).equals(
							((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getMyName()))
						{ 
							// Create a commited state with a commitTimeStamp
								Version v = ((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).newVersion();
								v.setCommitTime(commitTime);
								historyOfCommitedStates.add(v);
								System.out.println("	Commited @"+Thread.currentThread().getName()+" : <" 
								 + ((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getMyName() +"> CommitTime<"+ commitTime+">");
								 

							//Release write lock
							((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getLock().releaseWriteLock();
							System.out.println("[Release_Transaction_Write_Lock @ <"+ 
									((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
							
						}
					}
				}
			}
			
			for(int j = 0; j <  ((TransactionLocks)locks.get()).getUpdateLocks().size(); j++)
			{
				for(int i = 0; i < Tracked.getUpdatedObjects().size(); i++)
				{
					if(Tracked.getUpdatedObjects().elementAt(i) instanceof IMultiVersionLockBased )
					{
						if( ((String)((TransactionLocks)locks.get()).getUpdateLocks().elementAt(j)).equals(
							((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName()))
						{
							//Create a commited state with a commitTimeStamp
								Version v = ((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).newVersion();
								v.setCommitTime(commitTime);
								historyOfCommitedStates.add(v);
								System.out.println("	Commited @"+Thread.currentThread().getName()+" : <" 
										 + ((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName() +"> CommitTime<"+ commitTime+">");
										 
							// Release update lock
							 ((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getLock().releaseUpdateLock();
							 System.out.println("[Release_Transaction_Update_Lock @ <"+ 
										((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName() + "> For_Thread <" +Thread.currentThread().getName()+">" );
								
						}
					}
				}
			}
			
		}catch(NoZoneFoundForThisThreadException e)
		{
			e.printStackTrace();
		}
	}
	
	
	
	/**
	 *  Release transactional locks on transaction abort
	 */
	after() : signalMultiVersionTransactionAbort()
	{
	  try{
			for(int j = 0; j <  ((TransactionLocks)locks.get()).getWriteLocks().size(); j++)
			{
				for(int i = 0; i < Tracked.getWriteObjects().size(); i++)
				{
					if(Tracked.getWriteObjects().elementAt(i) instanceof IMultiVersionLockBased )
					{
						if( ((String)((TransactionLocks)locks.get()).getWriteLocks().elementAt(j)).equals(
							((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getMyName()))
						{
							System.out.println("[[ Release_Transaction_Write_Lock -> ON_OBJECT "+ 
									((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getMyName() + " -> FOR_THREAD " +Thread.currentThread().getName() );
							((IMultiVersionLockBased)Tracked.getWriteObjects().elementAt(i)).getLock().releaseWriteLock();
						}
					}
						
				}
			}
			
			for(int j = 0; j <  ((TransactionLocks)locks.get()).getUpdateLocks().size(); j++)
			{
				for(int i = 0; i < Tracked.getUpdatedObjects().size(); i++)
				{
					if(Tracked.getUpdatedObjects().elementAt(i) instanceof IMultiVersionLockBased )
					{
						if( ((String)((TransactionLocks)locks.get()).getUpdateLocks().elementAt(j)).equals(
							((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName()))
						{
							System.out.println("[[ Release_Transaction_Update_Lock -> ON_OBJECT "+ 
									((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getMyName() + " -> FOR_THREAD " +Thread.currentThread().getName() );
							((IMultiVersionLockBased)Tracked.getUpdatedObjects().elementAt(i)).getLock().releaseUpdateLock();
						}
					}
						
				}
			}
			
		}catch(NoZoneFoundForThisThreadException e)
		{
			e.printStackTrace();
		}
	}
	
	
	/**
	 * create a Read-only transaction with a unique creation time-stamp
	 */
	public static synchronized int startReadOnlyTransaction()
	{
		int startTime = -1;
		if(creationTime.get()== null)
		{
			startTime = counter.getNext();
			creationTime.set(new Integer(startTime));
		}
		
		return startTime;
	}
	
	
	/**
	 * Helper method to around advice
	 */
	private synchronized String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
	  

}
