/**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Provides Optimistic concurrency control with
 * 			backward validation to all objects of classes 
 * 			that implements the IOptimisticBackwardValidation
 * 			interface.
 */
package ca.mcgill.acid.aspects;

import ca.mcgill.acid.interfaces.ITracked;
import ca.mcgill.acid.interfaces.IShared;
import ca.mcgill.acid.interfaces.IOptimisticBackwardValidation;
import ca.mcgill.acid.interfaces.IAccessClassified;
import ca.mcgill.acid.interfaces.IAutoRecoverable;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import java.util.StringTokenizer;
import java.util.Vector;
import ca.mcgill.acid.interfaces.Constants;
import ca.mcgill.acid.interfaces.IVersioned;
import ca.mcgill.acid.utility.Version;
import ca.mcgill.acid.utility.NoZoneFoundForThisThreadException;
import ca.mcgill.acid.utility.GlobalCounter;
import ca.mcgill.acid.utility.ValidationFailedException;

public aspect OptimisticBackwardValidation {

	declare parents: 
		IOptimisticBackwardValidation implements IAccessClassified,IShared,IAutoRecoverable,ITracked;
	
	private int IOptimisticBackwardValidation.commitTime = 0;
    private void IOptimisticBackwardValidation.setCommitTime(int time)  { commitTime = time; }
    private int IOptimisticBackwardValidation.getCommitTime()  { return commitTime; }
    private static GlobalCounter counter = new GlobalCounter();
    private Vector deferredObjects = new Vector();
    
	/**
	 * Method call on the target object IOptimisticBackwardValidation
	 */
	pointcut methodCall(IOptimisticBackwardValidation optimistic): 
		target(optimistic) && call(public * IOptimisticBackwardValidation+.*(..));
	
	/**
	 * Joinpoints that signals transaction commit
	 */
	pointcut signalOptimisticBackwardTransactionCommit(): 
		execution(public static void commitOptimisticBackwardValidation());
	
	
	/**
	 * Only responsibility is to set the update strategy to
	 * Deferred for Write/Update operations
	 */
	Object around(IOptimisticBackwardValidation optimistic) 
	:  methodCall(optimistic)
	{	
		Object obj = null;
		String accessType = null;
		
		//get access type
		try{
			accessType = optimistic.getAccessType(getMethodName(thisJoinPoint.toShortString()));	
		}
		catch (MethodNotAnnotatedException e)
		{e.printStackTrace();}
		catch (MethodNotFoundException e)
		{e.printStackTrace();}
				
		// Set update strategy to Deferred for Write/Update operations
		if(!Constants.READ.equalsIgnoreCase(accessType)){
			optimistic.setDeferred(true);
		}
			
		obj = proceed(optimistic);
		
		return obj;
	}
	
	
	/**
	 * Validates the results of the commiting transaction and
	 * commit the results if successful
	 */
	before(): signalOptimisticBackwardTransactionCommit()
	{
		try{
			
			if(validationPhase())
			{ 
				System.out.println("	Successful Validation of Transaction <" +Thread.currentThread().getName()+ ">");
				writePhase();
			}
			else
				throw new ValidationFailedException("\n Validation was unsuccessful for the transaction <"
													+Thread.currentThread().getName()+ ">");	
		}catch(ValidationFailedException e){
			e.printStackTrace();
		}
		
    }
	
	
	/**
	 * -> Backward validation of commiting transaction <-
	 * For ALL modified object:
	 * 	If the commitTime of the deferred version equals the commitTime of the main version
	 * 		validation = successful
	 *  else
	 *  	validation = unsuccessful
	 */
	private synchronized boolean validationPhase()
	{
		System.out.println("	Begin Validation of Transaction <" +Thread.currentThread().getName()+ ">");
		boolean success = false;
		int versionID=-1;
		deferredObjects.clear();
		
		try{
			
			for(int i = 0; i < Tracked.getModifiedObjects().size(); i++)
			{
				if(Tracked.getModifiedObjects().elementAt(i) instanceof IOptimisticBackwardValidation )
				{
					//Get the versionID of the deferred version from AutoRecoverable
					for(int j = 0; j < AutoRecoverable.getObjects().size(); j++)
					{
						if( ((Version)AutoRecoverable.getObjects().elementAt(j)).getObjectName().equals(
							((IOptimisticBackwardValidation)Tracked.getModifiedObjects().elementAt(i)).getMyName()))
							{ 
								versionID = ((Version)AutoRecoverable.getObjects().elementAt(j)).getVersionID();
								break;
							}
					}
					
					// Compare the commitTime of the deferred vesion to that of the main version
					for(int k = 0; k < Versioned.getObjects().size(); k++ )
					 {
						  if(((IVersioned)Versioned.getObjects().elementAt(k)).getVersionID() == versionID
						 && ((IVersioned)Versioned.getObjects().elementAt(k)).getMyName().equals(
						    ((IOptimisticBackwardValidation)Tracked.getModifiedObjects().elementAt(i)).getMyName()))
						  {
							  if(((IOptimisticBackwardValidation)Tracked.getModifiedObjects().elementAt(i)).getCommitTime() ==
								  ((IOptimisticBackwardValidation)Versioned.getObjects().elementAt(k)).getCommitTime())
							  {
								  success = true;
								  deferredObjects.add(Versioned.getObjects().elementAt(k));
								  break;
							  }
							  else
								  return false;
						 }
					 }	
				}
			}
			
		}catch(NoZoneFoundForThisThreadException e){
			e.printStackTrace();
		}
		
		return success;
	}
	
	/*
	private synchronized boolean validationPhase(){
		Tbegin = ((Integer) (beginTime.get())).intValue();
		Tend = mostRecentCommitTime;
		Vector modifiedByOthers = new Vector();

		for(int i = 0; i < Tracked.getAllModifiedObjects().size(); i++){
		  if(Tracked.getAllModifiedObjects().elementAt(i) instanceof IOptimistic){
			if(((IOptimistic)Tracked.getAllModifiedObjects().elementAt(i)).getCommitTime() > Tbegin && 
			  ((IOptimistic)Tracked.getAllModifiedObjects().elementAt(i)).getCommitTime() <= Tend){
				modifiedByOthers.add(Tracked.getAllModifiedObjects().elementAt(i));
			}
		  }
		}
			
		for(int i = 0; i < Tracked.getModifiedObjects().size(); i++){
		if(modifiedByOthers.contains(Tracked.getModifiedObjects().elementAt(i))){
			return false;
		}
		}
		return true;
   }
*/
	
	
	/**
	 * Commit the changes of a transaction upon successful validation
	 */
	private void writePhase()
	{
		System.out.println("	Begin WritePhase of Transaction <" +Thread.currentThread().getName()+ ">");
		try
		{
			int commitTimeStamp = counter.getNext();
			
			//assign a new commit time stamp to all deferred objects and
			//replace the states of the main versions with that of the deferred versions
			for(int i = 0; i < Tracked.getModifiedObjects().size(); i++)
			{
				if(Tracked.getModifiedObjects().elementAt(i) instanceof IOptimisticBackwardValidation )
				{
					for(int k = 0; k < deferredObjects.size(); k++ )
					 {
						  if( ((IOptimisticBackwardValidation)deferredObjects.elementAt(k)).getMyName().equals(
						      ((IOptimisticBackwardValidation)Tracked.getModifiedObjects().elementAt(i)).getMyName()))
						  {
							((IOptimisticBackwardValidation)deferredObjects.elementAt(k)).setCommitTime(commitTimeStamp);
							((IOptimisticBackwardValidation)Tracked.getModifiedObjects().elementAt(i)).replaceState(deferredObjects.elementAt(k)); 
							break;
						  }
					 }	
					
					((IOptimisticBackwardValidation)Tracked.getModifiedObjects().elementAt(i)).setToMainVersion();
				}
				
			}
		}catch(NoZoneFoundForThisThreadException e){
			e.printStackTrace();
		} 
	
	}
	
	/**
	 * Helper method to around advice
	 */
	private synchronized String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
}
