 /**
 * @author Ekwa Duala-Ekoko
 * @version 1.0
 * Purpose: Keeps track of all the objects that are observed or
 * 			modified by a thread in a given zone.
 */
package ca.mcgill.acid.aspects;
 
import ca.mcgill.acid.interfaces.ITracked;
import ca.mcgill.acid.interfaces.INamed;
import ca.mcgill.acid.utility.Zone;
import java.util.Vector;
import java.util.StringTokenizer;
import ca.mcgill.acid.utility.MethodNotAnnotatedException;
import ca.mcgill.acid.utility.MethodNotFoundException;
import ca.mcgill.acid.interfaces.IAccessClassified;
import ca.mcgill.acid.utility.NoZoneFoundForThisThreadException;
import ca.mcgill.acid.interfaces.Constants;

 
public aspect Tracked {
   
	declare parents: ITracked implements INamed,IAccessClassified; 
 
	private static InheritableThreadLocal myTrackedZone = new InheritableThreadLocal();
	 	
	/**
     * Method call on the target object ITracked
     */
	pointcut transactionalMethodCall(ITracked track): 
		target(track) && call(public * ITracked+.*(..));
	
	//Keep track of object access
	after(ITracked track) :  transactionalMethodCall(track) && if(myTrackedZone.get()!= null)
	{
		String type = "";
			try{
				//Get access type
				type = track.getAccessType(getMethodName(thisJoinPoint.toShortString()));
			}
			catch (MethodNotAnnotatedException e)
			{e.printStackTrace();}
			catch (MethodNotFoundException e)
			{e.printStackTrace();}
			
			//Place the target object in the appropriate category
			if(type.equalsIgnoreCase(Constants.READ))
			{	
				if(!zoneContainsObject(((Zone)myTrackedZone.get()).getReadObjects(),(ITracked)thisJoinPoint.getTarget()))
					((Zone)myTrackedZone.get()).addReadObject(thisJoinPoint.getTarget());				
			}	
			
			if(type.equalsIgnoreCase(Constants.WRITE))
			{
				if(!zoneContainsObject(((Zone)myTrackedZone.get()).getWriteObjects(),(ITracked)thisJoinPoint.getTarget()))
					((Zone)myTrackedZone.get()).addWriteObject(thisJoinPoint.getTarget());				
			}	

			if(type.equalsIgnoreCase(Constants.UPDATE))
			{
				if(!zoneContainsObject(((Zone)myTrackedZone.get()).getUpdatedObjects(),(ITracked)thisJoinPoint.getTarget()))
					((Zone)myTrackedZone.get()).addUpdatedObject(thisJoinPoint.getTarget());				
			}	
			
			if(type.equalsIgnoreCase(Constants.WRITE) || type.equalsIgnoreCase(Constants.UPDATE))
			{
				if(!zoneContainsObject(((Zone)myTrackedZone.get()).getModifiedObjects(),(ITracked)thisJoinPoint.getTarget()))	
					((Zone)myTrackedZone.get()).addModifiedObject(thisJoinPoint.getTarget());
			}
			
	}
	
	
	/**
	 * Begin a new track zone for a thread
	 */
	public static synchronized void beginTrackedZone()
	{
		if(myTrackedZone.get()== null)
		{
			Zone myZone = new Zone();
			myTrackedZone.set(myZone);
		}
	}
	
	/**
	 * End an existing track zone
	 */
	public static synchronized void endTrackedZone()
	{
		 myTrackedZone.set(null);
	}
	
	
	/**
	 * Join an existing track zone
	 */
	public static synchronized void joinTrackedZone(Zone zone)
	{
		 myTrackedZone.set(zone);
	}
	
	/**
	 * Leave an existing track zone
	 */
	public static synchronized void leaveTrackedZone()
	{
		 myTrackedZone.set(null);
	}
	
	/**
	 * returns the track zone of the calling thread
	 */
	public static synchronized Zone getMyTrackedZone()
	{
		return (Zone)myTrackedZone.get();
	}
	
	/**
	 * returns the objects read by the invoking thread
	 */
	public static synchronized Vector getReadObjects()throws NoZoneFoundForThisThreadException
	{	
		 if(myTrackedZone.get()== null)
			throw new NoZoneFoundForThisThreadException(
					"\n	No Tracked Zone found for Thread: <" 
					+ Thread.currentThread().getName() + ">" );
		
		return ((Zone)myTrackedZone.get()).getReadObjects();
	}
	
	/**
	 * returns the objects written-to by the invoking thread
	 */
	public static synchronized Vector getWriteObjects()throws NoZoneFoundForThisThreadException
	{	
		 if(myTrackedZone.get()== null)
			throw new NoZoneFoundForThisThreadException(
					"\n	No Tracked Zone found for Thread: <" 
					+ Thread.currentThread().getName() + ">" );
		
		return ((Zone)myTrackedZone.get()).getWriteObjects();
	}
	
	/**
	 * returns the objects updated by the invoking thread
	 */
	public static synchronized Vector getUpdatedObjects()throws NoZoneFoundForThisThreadException
	{	
		 if(myTrackedZone.get()== null)
			throw new NoZoneFoundForThisThreadException(
					"\n	No Tracked Zone found for Thread: <" 
					+ Thread.currentThread().getName() + ">" );
		
		return ((Zone)myTrackedZone.get()).getUpdatedObjects();
	}

	/**
	 * returns the objects modified by the invoking thread
	 */
	public static synchronized Vector getModifiedObjects()throws NoZoneFoundForThisThreadException
	{

		 if(myTrackedZone.get()== null)
				throw new NoZoneFoundForThisThreadException(
						"\n	No Tracked Zone found for Thread: <" 
						+ Thread.currentThread().getName() + ">" );
		return ((Zone)myTrackedZone.get()).getModifiedObjects();
	}
	
	/**
	 * Helper method
	 */
	private String getMethodName(String s)
	{

		StringTokenizer st = new StringTokenizer(s,"."); 
		String temp = st.nextToken();
		
		StringTokenizer st2 = new StringTokenizer(st.nextToken(), "(");
		
		return (st2.nextToken()).trim();
	}
	  
	/**
	 * Helper method
	 */
	private boolean zoneContainsObject(Vector zone, ITracked obj)
	{
		boolean found = false;
		for(int i = 0; i < zone.size() ; i++)
		{
			if(obj.getMyName().equalsIgnoreCase(((ITracked)zone.elementAt(i)).getMyName()))
			{
				found = true;
				break;
				
			}
		}
		
		return found;
	}
	
}
